# the bit that listens idly and serves RPCs
# and executor is the thing that takes a function, and executes it in a thread/process and returns the result as response or even another HTTP req entirely

from .utils import deserialize, serialize, get_self_ip, overwrite
from .comms_wrappers import simplex_wrapper, duplex_wrapper
from .config import comms_config, default_rpc_config, default_service_config, default_service_depth
from .inline_executor import InlineExecutor

from flask import Flask
from flask import request as route_req
from multiprocessing import Process
from copy import copy
import threading
import inspect
import random
import string

# where the services being offered are stored
rpcs = []
# the ip address of the worker
ip_addr = get_self_ip()
# the app that listens for incomming http requests
app = Flask(__name__)
name = 'worker'
node_type = 'worker'

# a default route to serve up what rpcs this worker offers, and their configuration
@app.route('/_get_profile', methods=['GET'])
def _get_profile():
	global name, ip_addr, rpcs

	profile = {
		'name': name,
		'ip_addr': ip_addr,
		'rpcs': RPC_node.get_profile()
	}

	# print('----------------------------------------------------------------')
	# print(profile['rpcs']['simplex_rpc']['__call__']['comms_pattern'])
	# print(profile['rpcs']['duplex_rpc']['__call__']['comms_pattern'])
	# print('----------------------------------------------------------------')

	return serialize(profile)

# a default route to kill the worker in case a bug blocks SIGINT
@app.route('/_kill', methods=['GET'])
def kill():
	func = route_req.environ.get('werkzeug.server.shutdown')

	if func is None:
		raise RuntimeError('Not running with the Werkzeug Server')

	func()
	return 'shutting down'

# a default route to provide basic info about the axon node, namely, that it's a worker
@app.route('/_type', methods=['GET'])
def _type():
	return node_type

# this returns the function that actually gets registered as a route in flask.
def get_route_fn(configuration, fn):
	# wraps fn in two layers
	# the outer layer handles the communication pattern, being either simplex or duplex
	# the inner layer handles the execution environment, be that a thread or a process

	comms_pattern = configuration['comms_pattern']
	comms_wrapper = None
	if (comms_pattern == 'simplex'):
		comms_wrapper = simplex_wrapper
	elif (comms_pattern == 'duplex'):
		comms_wrapper = duplex_wrapper
	else:
		raise BaseException('unrecognized comms_pattern: '+str(comms_pattern))

	try:
		executor = configuration['executor']
	except(KeyError):
		raise(KeyError('executor not specified in RPC configuration'))

	if (executor == 'inline'):
		executor = InlineExecutor
	elif (executor == 'Process'):
		executor = Process
	elif((executor == 'Thread')):
		executor = threading.Thread
	else:
		raise(BaseException('unrecognized executor configuration:', executor))

	wrapped_fn = comms_wrapper(fn, executor)

	return wrapped_fn

# this function returns a function which will be applied to functions which the caller wishes to turn into RPCs
def make_RPC_skeleton(**configuration):
	# accepts options for the RPC, simplex/duplex, and running environment
	# sets up the function that registers the route

	# overwrites the default configuration with keys from caller input
	configuration = overwrite(default_rpc_config, configuration)

	# this is the function that will be applied to functions which the caller wishes to turn into RPCs
	# accepts a function, which it will wrap in an executor function depending on the configuration passed to rcp, this wrapped function will be registered as a route
	def init_rpc(fn):

		if not 'name' in configuration:
			configuration['name'] = fn.__name__
		
		rpcs.append(configuration)

		# wraps fn in the needed code to deserialize parameters from the request return its results, as well as run in a separate process/thread
		route_fn = get_route_fn(configuration, fn)
		# functions that flask registers as routes must have different names, even if they are registered at different endpoints. get_executor returns a function called 'wrapped_fn' every time, so we've got to change the name of the function here.
		# it is set to a random string firstly because the literal name of the function is arbitrary, but also because some services might have functions with the same names
		route_fn.__name__ = ''.join(random.choices(string.ascii_letters, k=10))
		# the endpoint of the route which exposes the function
		endpoint = '/'+configuration['endpoint_prefix']+configuration['name']

		# registering the function as a route
		app.route(endpoint, methods=['POST'])(route_fn)

		return fn

	return init_rpc

class ServiceNode():

	def __init__(self, subject, name, depth=default_service_depth, **configuration):
		self.subject = subject
		self.name = name
		self.children = {}
		self.depth = depth

		self.configuration = overwrite(default_service_config, configuration)

		# iterates over members and either registers them as RPCs or recursively turns them into ServiceNodes
		# for key, member in self.subject.__dict__.items():
		for key in dir(self.subject):

			# dir calls the overwritable __dir__ function on self.subject. This means it's unreliable in some cases, since developers can specify what it returns.
			# this can mean the attribute is listed but does not exist, or even that calling the attribute throws an error
			try:
				if not hasattr(self.subject, key):
					continue

			except(BaseException):
				# sometimes merely accessing an attribute can throw an error
				continue

			member = getattr(self.subject, key)

			if (key == '__call__'):
				# Any member accessed via __call__ attribute is represented in profile by an RPC config. This means that objects stored on __call__ attributes will not be represented in profile
				self.init_RPC(key, member)
			
			elif hasattr(member, '__dict__'):
				# Any member with a __dict__ attribute gets a profile, provided it as not been accessed via __call__
				self.add_child(key, member)

			elif hasattr(member, '__call__'):
				# Any member with a __call__ attribute but no __dict__ attribute is represented in profile by an RPC config
				self.init_RPC(key, member)

		# we now register a GET route at the ServiceNode's endpoint to expose its profile
		
		# the function that will be exposed as a route
		def get_profile_str():
			profile = self.get_profile()
			return serialize(profile)

		# functions that flask registers as routes must have different names, even if they are registered at different endpoints, so we've got to change the name of the function here.
		# it is set to a random string firstly because the literal name of the function is arbitrary, but also because some services might have functions with the same names
		get_profile_str.__name__ = ''.join(random.choices(string.ascii_letters, k=10))
		endpoint = '/'+self.configuration['endpoint_prefix']+self.name

		app.route(endpoint, methods=['GET'])(get_profile_str)

	def add_child(self, key, child, **child_config):
		# limits recursion to a depth parameter
		if (self.depth <= 0): return

		# The child config overwrites the parent's config, meaning by default children inherit configuration from their parents
		child_config = overwrite(self.configuration, child_config)

		# the child's endpoint_prefix is the parent's prefix plus / and the child's name
		child_config['endpoint_prefix'] += str(self.name)+'/'

		# create a ServiceNode out of it and register it as a child
		child = ServiceNode(child, key, depth=self.depth-1, **child_config)
		self.children[key] = child

	def init_RPC(self, key, fn):
		child_config = copy(self.configuration)
		child_config['endpoint_prefix'] += str(self.name)+'/'
		child_config['name'] = key

		# make it an RPC
		make_rpc = make_RPC_skeleton(**child_config)
		make_rpc(fn)
		# remember the configuration
		self.children[key] = child_config

	# returns a JSON serializable dict tree with leaves of RPC configuration dicts
	def get_profile(self):
		
		profile = {}

		# this is a flag attribute that lets a ServiceStub know that a dict is a profile for a ServiceNode, not a configuration dict for an RPC
		profile['__profile_flag__'] = True

		for key in self.children.keys():
			child = self.children[key]

			# if the child is itself a ServiceNode, recursively get the profiles of its attributes
			if isinstance(child, ServiceNode):
				profile[key] = child.get_profile()

			# if the attribute is an RPC, the profile element is its configuration
			else:
				profile[key] = child

		return profile

# a ServiceNode that holds all RPCs associated by this worker instance
RPC_node = ServiceNode(object(), 'rpcs')

# the RPC decorator adds the associated function to the RPC_node ServiceNode
def rpc(**configuration):
	def add_to_RPC_node(fn):
		RPC_node.add_child(fn.__name__, fn, **configuration)

	return add_to_RPC_node

# starts the web app
def init(wrkr_name='worker'):
	global name

	name = wrkr_name
	# the web application that will serve the services and rpcs as routes
	app.run(host='0.0.0.0', port=comms_config.worker_port, threaded=False)


from threading import Lock

# this is so that no other inline functions will execute when an inline duplex RPC is running
# Inline duplex RPCs need to run in a separate thread because of WSGI, which mandates that a route stops executing once it has sent its response, preventing duplex RPCs from sending a return POST
inline_lock = Lock()

class InlineExecutor():

	def __init__(self, target=lambda x : x, args=('null',)):
		self.target = target
		self.args = args

	def start(self):
		with inline_lock:
			self.target(*self.args)

	def join(self):
		pass


from .utils import overwrite
from .config import default_service_config, default_rpc_config
from .worker import rpc

# This is possibly a piece of lint, service_client_tests.py does not import it and it hasn't been kept curent as it does not check for errors when accessing attributes

class ServiceNode():

	def __init__(self, subject, name, **configuration):

		self.subject = subject
		self.name = name
		self.configuration = configuration
		self.children = {}

		configuration = overwrite(default_service_config, configuration)

		# iterates over members and either registers them as RPCs or recursively turns them into ServiceNodes
		# for key, member in self.subject.__dict__.items():
		for key in dir(self.subject):
			member = getattr(self.subject, key)

			# if the member is callable, make it an RPC
			if callable(member):
				self.init_RPC(member, configuration)

			# this is for the case when the object is itself an object, not a function or a primative type
			elif hasattr(member, '__dict__'):
				self.init_child(key, member, configuration)

	def init_child(self, key, member, configuration):
		child_config = overwrite(configuration, {'endpoint_prefix': configuration['endpoint_prefix']+key+'/'})
		child = ServiceNode(member, key, **child_config)
		self.children[key] = child

	def init_RPC(self, fn, configuration):
		# make it an RPC
		make_rpc = rpc(**configuration)
		make_rpc(fn)
		# remember the configuration
		self.children[key] = configuration

	# returns a JSON serializable dict tree with leaves of RPC configuration dicts
	def get_profile(self):
		
		profile = {}

		for key in self.children.keys():
			child = self.children[key]

			# if the child is itself a ServiceNode, recursively get the profiles of its attributes
			if isinstance(child, ServiceNode):
				profile[key] = child.get_profile()

			# if the attribute is an RPC, the profile element is its configuration
			else:
				profile[key] = child

		return profile


from .utils import serialize, deserialize, GET, POST, async_GET, async_POST
from .config import comms_config, default_rpc_config
from .return_value_linker import ReturnEvent_async, ReturnEvent_coro, RVL
from .simplex_stubs import error_handler, AsyncCallHandle, GenericSimplexStub

import threading
import asyncio
import uuid

# the Return Value Linker is an app that runs in a separate thread and listens for incomming HTTP requests carrying the results from duplex RPC calls
rvl = RVL()

# this function checks to see if the RVL ap is running, and if not, starts it
def ensure_rvl_app():
	global rvl

	if not rvl.running:
		rvl.start_app()

# this function ensures the RVL is active and listening for the return value and then makes a calling request to a duplex RPC
def call_duplex_rpc_async(url, args, kwargs):
	global rvl

	# checks that the RVL is active, and if not, starts one
	ensure_rvl_app()

	return_event = ReturnEvent_async()

	def thread_fn(return_event):
		# we must register an event listenner with the return value linker, which will wait for the incomming result request
		# this uuid will identify the call, so the RVL can lookup the event
		call_id = uuid.uuid4()
		rvl.register(call_id, return_event)

		# this object holds information about the function call that the worker will need to provide to the rvl upon completion
		call_info = {'id': call_id, 'rvl_port': rvl.port}

		# makes the calling request
		status, text = POST(url=url , data={'msg': serialize((call_info, args, kwargs))})
		return_obj = deserialize(text)

		# raises exception if the receiving RPC is simplex, not duplex
		if (return_obj != 'duplex'):
			print(return_obj)
			raise(BaseException('duplex call sent to simplex RPC'))

	request_thread = threading.Thread(target=thread_fn, args=(return_event, ), name='call_duplex_rpc_async.request_thread')
	request_thread.start()

	return AsyncCallHandle(return_event)

# this function ensures the RVL is active and listening for the return value and then makes a calling request to a duplex RPC
async def call_duplex_rpc_coro(url, args, kwargs):
	global rvl
	
	ensure_rvl_app()

	# we must register an event listenner with the return value linker, which will wait for the incomming result request
	# this uuid will identify the call, so the RVL can lookup the event
	call_id = uuid.uuid4()
	return_event = ReturnEvent_coro()
	await return_event.init()
	rvl.register(call_id, return_event)

	# this object holds information about the function call that the worker will need to provide to the rvl upon completion
	call_info = {'id': call_id, 'rvl_port': rvl.port}

	# makes the calling request
	status, text = await async_POST(url=url , data={'msg': serialize((call_info, args, kwargs))})
	return_obj = deserialize(text)

	# raises exception if the receiving RPC is simplex, not duplex
	if (return_obj != 'duplex'):
		print(return_obj)
		raise(BaseException('duplex call sent to simplex RPC'))

	# we now wait for the result to return
	serialized_result = await return_event.get_return_value()

	# deserializes result
	result = deserialize(serialized_result)

	# detects if an error was thrown in worker
	return error_handler(result)

# this function ensures the RVL is active and listening for the return value and then makes a calling request to a duplex RPC
def call_duplex_rpc_sync(url, args, kwargs):
	global rvl
	
	ensure_rvl_app()

	# we must register an event listenner with the return value linker, which will wait for the incomming result request
	# this uuid will identify the call, so the RVL can lookup the event
	call_id = uuid.uuid4()
	return_event = ReturnEvent_async()
	rvl.register(call_id, return_event)

	# this object holds information about the function call that the worker will need to provide to the rvl upon completion
	call_info = {'id': call_id, 'rvl_port': rvl.port}

	# makes the calling request
	status, text = POST(url=url , data={'msg': serialize((call_info, args, kwargs))})
	return_obj = deserialize(text)

	# raises exception if the receiving RPC is simplex, not duplex
	if (return_obj != 'duplex'):
		print(return_obj)
		raise(BaseException('duplex call sent to simplex RPC'))

	# we now wait for the result to return
	serialized_result = return_event.get_return_value()

	# deserializes result
	result = deserialize(serialized_result)

	# detects if an error was thrown in worker
	return error_handler(result)

class GenericDuplexStub(GenericSimplexStub):

	def __init__(self, *args, **kwargs):
		GenericSimplexStub.__init__(self, *args, **kwargs)

	def async_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__
		
		return call_duplex_rpc_async(url, args, kwargs)

	async def coro_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		return await call_duplex_rpc_coro(url, args, kwargs)

	def sync_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		return call_duplex_rpc_sync(url, args, kwargs)

class SyncDuplexStub(GenericDuplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.sync_call(args, kwargs)

class AsyncDuplexStub(GenericDuplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.async_call(args, kwargs)

class CoroDuplexStub(GenericDuplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	async def __call__(self, *args, **kwargs):
		return await self.coro_call(args, kwargs)


from .config import comms_config
from .utils import serialize, deserialize

from flask import Flask
from flask import request as route_req

# this class is a notice board, which workers can sign into and out of to notify other workers of their presence
class NoticeBoard():

	def __init__(self, name='notice board', port=comms_config.notice_board_port, show_logs=True):

		self.ip_set = set({})
		self.app = Flask(name)
		self.port = port
		self.show_logs = show_logs

		# we now add routes to the app

		@self.app.route('/sign_in', methods=['GET'])
		def sign_in():
			remote_addr = route_req.remote_addr
			self.ip_set.add(remote_addr)

			self.log('/sign_in: ip address:', remote_addr, 'recorded, now have', len(self.ip_set), 'registered addresses')

			return 'sign in successful'

		@self.app.route('/sign_out', methods=['GET'])
		def sign_out():
			remote_addr = route_req.remote_addr

			try:
				self.ip_set.remove(remote_addr)
				self.log('/sign_out: ip address:', remote_addr, 'removed, now have', len(self.ip_set), 'registered addresses')
				return 'sign out successful'

			except(KeyError):
				self.log('/sign_out: ip address:', remote_addr, 'not recorded')
				return 'ip address not recorded'


		@self.app.route('/get_ips', methods=['GET'])
		def get_ips():
			self.ip_list = list(self.ip_set)
			return serialize(self.ip_list)

		# route to shut the server down in case sigint fails
		@self.app.route('/kill')
		def kill():
			func = route_req.environ.get('werkzeug.server.shutdown')

			if func is None:
				raise RuntimeError('Not running with the Werkzeug Server')

			func()
			return 'shutting down'

		# a default route to provide basic info about the axon node, namely, that it's a notice board
		@self.app.route('/_type', methods=['GET'])
		def _type():
			return 'notice_board'

	def get_ips(self):
		return list(self.ip_set)

	def start(self):
		self.log('starting notice board app')
		self.app.run(host='0.0.0.0', port=self.port)

	def log(self, *args):
		if self.show_logs: print(*args)


from .utils import GET, POST
from .config import comms_config
from .utils import serialize, deserialize, get_self_ip
from .notice_board_app import NoticeBoard

from flask import Flask
from flask import request as route_req
import requests, aiohttp
import asyncio 

# broadcasts to all IPs on the LAN, and returns a list of the IPs that provide a response
async def broadcast_discovery(node_type='*', port=comms_config.worker_port, num_hosts=-1, timeout=10):
	self_ip = get_self_ip()
	# the first three numbers of the ip address, representing the LAN address
	base_addr = '.'.join(self_ip.split('.')[:-1])+'.'
	# the urls we will be sending requests to
	host_ips = range(0, 255)

	# the future that holds all the request coroutines
	all_reqs = None
	# where the IP addresses of the discovered nodes will be stored
	result_holder = {'result': []}

	# we need to create a clientsession to send async requests
	async with aiohttp.ClientSession(conn_timeout=timeout) as session:

		# coroutine that we will call on each ip in host_ips
		async def req_fn(ip, result_holder):
			url = 'http://'+base_addr+str(ip)+':'+str(port)+'/_type'
			status = None
			test = None

			# catches no connection and timeout errors
			try:
				# sends request and awaits result
				async with session.get(url) as resp:
					status = resp.status
					text = await resp.text()

				# if the host at ip exposes the endpoint we're looking for, the response status should be in the 200s
				if (200 <= status < 300):
					# checks to make sure that the found node is the right kind of node, if the caller provides a type of node
					if (node_type == '*') or (text == node_type):

						result_holder['result'].append(base_addr+str(ip))

						# if we are to stop after some number of hosts, cancel the other requests
						if (len(result_holder['result']) >= num_hosts) and (0 < num_hosts):
							all_reqs.cancel()

				return

			except(aiohttp.client_exceptions.ClientConnectorError):
				return

			except(aiohttp.client_exceptions.ServerTimeoutError):
				return

		# the futures returned by sending requests to each host
		all_reqs = asyncio.gather(*[req_fn(ip, result_holder) for ip in host_ips])

		try:	
			await all_reqs

		except(asyncio.CancelledError):
			pass

	return result_holder['result']

def sign_in(ip='localhost', port=comms_config.notice_board_port):
	status, text = GET('http://'+str(ip)+':'+str(port)+'/sign_in')

	if (hash(text) != hash('sign in successful')):
		print('sign_in: notice board gave warning, may not be registered')

def sign_out(ip='localhost', port=comms_config.notice_board_port):
	status, text = GET('http://'+str(ip)+':'+str(port)+'/sign_out')

	if (hash(text) == hash('ip address not recorded')):
		print('sing_out: not registered at given ip')

	elif (hash(text) != hash('sign out successful')):
		print('sing_out: notice board gave warning')

def get_ips(ip='localhost', port=comms_config.notice_board_port):
	status, text = GET('http://'+str(ip)+':'+str(port)+'/get_ips')
	return deserialize(text)


from .utils import serialize, deserialize, GET, POST, async_GET, async_POST
from .config import comms_config, default_rpc_config
from .return_value_linker import ReturnEvent_async

import threading

# this function checks if an error flag has been set and raises the corresponding error if it has
def error_handler(return_obj):
	if (return_obj['errcode'] == 1):
		# an error occured in worker, raise it
		(error_info, error) = return_obj['result']

		print('the following error occured in worker:')
		print(error_info)
		raise(error)

	else:
		# returns the result
		return return_obj['result']

# this function makes a calling request to a simplex RPC
async def call_simplex_rpc_coro(url, args, kwargs):
	
	# makes the calling request
	status, text = await async_POST(url=url , data={'msg': serialize((args, kwargs))})

	if (text =='duplex'):
		raise(BaseException('simplex call sent to duplex RPC'))

	# deserializes return object from worker
	return_obj = deserialize(text)
	return error_handler(return_obj)

class AsyncCallHandle():

	def __init__(self, response_event):
		self.response_event = response_event

	def join(self):
		text = self.response_event.get_return_value()

		if (text =='duplex'):
			raise(BaseException('simplex call sent to duplex RPC'))

		# deserializes return object from worker
		return_obj = deserialize(text)
		return error_handler(return_obj)

# this function makes a calling request to a simplex RPC, and returns a handle which will block and return the result of the request on .join
def call_simplex_rpc_async(url, args, kwargs):

	response_event = ReturnEvent_async()
	
	def thread_fn(response_event):
		# makes the calling request
		status, text = POST(url=url , data={'msg': serialize((args, kwargs))})
		response_event.put_return_value(text)

	request_thread = threading.Thread(target=thread_fn, args=(response_event, ), name='call_duplex_rpc_async.request_thread')
	request_thread.start()

	return AsyncCallHandle(response_event)

# this function makes a calling request to a simplex RPC
def call_simplex_rpc_sync(url, args, kwargs):
	status, text = POST(url=url , data={'msg': serialize((args, kwargs))})
	
	if (text =='duplex'):
		raise(BaseException('simplex call sent to duplex RPC'))

	# deserializes return object from worker
	# print(text)
	return_obj = deserialize(text)
	return error_handler(return_obj)

class GenericSimplexStub:

	def __init__(self, worker_ip='localhost', rpc_name=None, endpoint_prefix=default_rpc_config['endpoint_prefix']):
		self.remote_ip = worker_ip
		self.__name__ = rpc_name
		self.endpoint_prefix = endpoint_prefix

	def async_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__
		
		return call_simplex_rpc_async(url, args, kwargs)

	async def coro_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		return await call_simplex_rpc_coro(url, args, kwargs)

	def sync_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		return call_simplex_rpc_sync(url, args, kwargs)

	# returns a boolean indicating if the stub has all the information it needs to make calls or not
	def check_capability(self):
		if (self.__name__ == None):
			return False

		else:
			return True

class AsyncSimplexStub(GenericSimplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.async_call(args, kwargs)

class CoroSimplexStub(GenericSimplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	async def __call__(self, *args, **kwargs):
		return await self.coro_call(args, kwargs)

class SyncSimplexStub(GenericSimplexStub):

	def __init__(self, **kwargs):
		GenericSimplexStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.sync_call(args, kwargs)


from .utils import serialize, deserialize, get_open_port
from .config import comms_config

from flask import Flask
from flask import request as route_req
from threading import Thread, Lock
import sys
import uuid
import time
import requests
import asyncio

# stops flask from outputting starting message
cli = sys.modules['flask.cli']
cli.show_server_banner = lambda *x: 'hello there'

# stops flask from printing to console when it fulfills a request
import logging
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)
log.disabled = True

# this class works the same way as asyncio events, except it cannot be cleared
# the important difference is that the arguements(s) given to set will be returned by await
class ReturnEvent_coro():

	def __init__(self):
		self.event = asyncio.Event()
		self.value = None
		self.event_loop = None

	async def init(self, loop=None):
		if loop == None:
			self.event_loop = asyncio.get_running_loop()
		else:
			self.event_loop = loop

	async def get_return_value(self):
		await self.event.wait()
		return self.value

	def put_return_value(self, value):

		# checks that the event loop exists and is running
		if self.event_loop == None:
			print('call ID:', uuid)
			raise(BaseException('ReturnEvent_coro not initialized with event loop'))
		elif not self.event_loop.is_running():
			print('call ID:', uuid)
			raise(BaseException('ReturnEvent_coro has inactive event loop'))

		# schedules a callback on the event loop that initialized this ReturnEvent_coro to set the return value

		def callback():
			self.value = value
			self.event.set()

		self.event_loop.call_soon_threadsafe(callback)

class ReturnEvent_async():

	def __init__(self):
		self.lock = Lock()
		self.lock.acquire()
		self.value = None

	# this function is called from a separate thread that wishes to get the value
	def get_return_value(self):
		self.lock.acquire()
		return self.value

	def put_return_value(self, value):
		
		self.value = value
		self.lock.release()

# this class encapsulates an app that will listen for incoming result requests from duplex RPCs
class RVL():

	def __init__(self):
		# the app that will run in a separate thread and listen for incomming result requests
		self.app = Flask(__name__)
		# boolean indiating weathre or not the app thread is running
		self.running = False
		# where asyncio events and such will be posted
		self.event_loop = None

		self.stubs = {}
		self.port = None

		# adds a route that will listen for incoming result requests from duplex RPCs
		def lookup():

			# the ID and result of the function call
			serialized_result = route_req.form['result']
			uuid = deserialize(route_req.form['id'])
			# gets the return event, with which we may pass the return value to the caller of the function
			return_event = self.stubs[uuid]
			# passes the return value back to the caller of the RPC
			return_event.put_return_value(serialized_result)
			# clears the record of the RPC call to prevent memory leak
			del self.stubs[uuid]

			return 'result returned successfully'

		self.app.route("/_return_value_linker", methods=['POST'])(lookup)

	def register(self, uuid, event):
		self.stubs[uuid] = event

	def start_app(self):

		def app_fn(port):
			self.app.run(host='0.0.0.0', port=port, threaded=False)

		# finds an available port
		rvl_port = get_open_port(lower_bound=comms_config.RVL_port)
		self.port = rvl_port
		# starts a thread that the app will run in
		self.app_thread = Thread(target=app_fn, args=(rvl_port,), name='rvl_thread')
		self.app_thread.daemon = True
		self.app_thread.start()

		# we now poll the rvl app until it starts responding to tell when it has started. We don't want the function to complete without anything listenning for incoming results

		# registers a return event
		call_id = uuid.uuid4()
		return_event = ReturnEvent_async()
		self.register(call_id, return_event)

		# this has to be done before this function completes or lookup will raise an exception
		self.running = True

		# polls every 0.1 seconds until we can set the event
		while True:
			time.sleep(0.1)

			try:
				url = 'http://localhost:'+str(rvl_port)+'/_return_value_linker'
				data = {'msg': serialize((call_id, 'init return val'))}

				# if the rvl port isn't open, this should throw
				requests.post(url=url, data=data)

				# if PC gets here, the RVL port is open, can stop polling
				break

			except(requests.exceptions.ConnectionError):
				pass



from .config import comms_config, default_rpc_config
from .simplex_stubs import call_simplex_rpc_coro, call_simplex_rpc_async, call_simplex_rpc_sync
from .duplex_stubs import call_duplex_rpc_coro, call_duplex_rpc_async, call_duplex_rpc_sync

class GenericStub():

	def __init__(self, worker_ip='localhost', rpc_name=None, endpoint_prefix=default_rpc_config['endpoint_prefix'], comms_pattern='simplex'):
		self.remote_ip = worker_ip
		self.__name__ = rpc_name
		self.endpoint_prefix = endpoint_prefix
		self.comms_pattern = comms_pattern

	def async_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__
		
		result = None

		if (self.comms_pattern == 'simplex'):
			result = call_simplex_rpc_async(url, args, kwargs)
		elif (self.comms_pattern == 'duplex'):
			result = call_duplex_rpc_async(url, args, kwargs)
		else:
			raise BaseException('unrecognized comms_pattern')

		return result

	async def coro_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		result = None
		if (self.comms_pattern == 'simplex'):
			result = await call_simplex_rpc_coro(url, args, kwargs)
		elif (self.comms_pattern == 'duplex'):
			result = await call_duplex_rpc_coro(url, args, kwargs)
		else:
			raise BaseException('unrecognized comms_pattern')

		return result

	def sync_call(self, args, kwargs):
		if not self.check_capability():
			raise(BaseException('stub not initialized'))

		url = 'http://'+str(self.remote_ip)+':'+str(comms_config.worker_port)+'/'+self.endpoint_prefix+self.__name__

		result = None

		if (self.comms_pattern == 'simplex'):
			result = call_simplex_rpc_sync(url, args, kwargs)
		elif (self.comms_pattern == 'duplex'):
			result = call_duplex_rpc_sync(url, args, kwargs)
		else:
			raise BaseException('unrecognized comms_pattern')

		return result

	# returns a boolean indicating if the stub has all the information it needs to make calls or not
	def check_capability(self):
		if (self.__name__ == None):
			return False

		else:
			return True

class AsyncStub(GenericStub):

	def __init__(self, **kwargs):
		GenericStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.async_call(args, kwargs)

class CoroStub(GenericStub):

	def __init__(self, **kwargs):
		GenericStub.__init__(self, **kwargs)

	async def __call__(self, *args, **kwargs):
		return await self.coro_call(args, kwargs)

class SyncStub(GenericStub):

	def __init__(self, **kwargs):
		GenericStub.__init__(self, **kwargs)

	def __call__(self, *args, **kwargs):
		return self.sync_call(args, kwargs)


# the bit that exports worker profile class and uses the RVL
# an RPC stub is the thing on the client that makes a calling request and waits for the response

from .config import comms_config, default_service_config
from .utils import deserialize, GET
from .simplex_stubs import AsyncSimplexStub, CoroSimplexStub, SyncSimplexStub
from .duplex_stubs import AsyncDuplexStub, CoroDuplexStub, SyncDuplexStub
from .generic_stubs import AsyncStub, CoroStub, SyncStub

from types import SimpleNamespace

def get_ServiceStub(ip_addr='localhost', endpoint_prefix=default_service_config['endpoint_prefix'], name='', profile=None, stub_type=CoroStub, top_stub_type=object):
	global count 

	# the attributes of the returned object
	attrs = {}

	# gets the profile if none are provided
	if (profile == None):
		url = 'http://'+str(ip_addr)+':'+str(comms_config.worker_port)+'/'+endpoint_prefix+name
		_, profile_str = GET(url)
		profile = deserialize(profile_str)

	# once the profil is obtained, metaclass creation is left to get_ServiceStub_helper
	return get_ServiceStub_helper(ip_addr, profile, stub_type, top_stub_type)

def get_ServiceStub_helper(ip_addr, profile, stub_type, top_stub_type):

	attrs = {}
	keys = list(profile.keys())
	banned_keys = ['__call__', '__profile_flag__', '__func__', '__self__', '__get__', '__set__', '__delete__'] + dir(object())

	for key in keys:

		if (key in banned_keys): continue

		member = profile[key]

		if '__profile_flag__' in member:
			# member is a profile for a ServiceNode
			attrs[key] = get_ServiceStub_helper(ip_addr, member, stub_type, object)

		else:
			# If a member is not a profile, then it must be an RPC config, and so correspond to a callable object on the worker with no __dict__attribute
			attrs[key] = stub_type(worker_ip=ip_addr, endpoint_prefix=member['endpoint_prefix']+'/', rpc_name=key)

	parent_classes = (top_stub_type, )

	if '__call__' in keys:
		# if the profile has a __call__ attribute, than the corresponding object on the server is callable and has a __dict__ attribute, and so must be represented by an RPC stub bound to the given network coordinates
		BoundStubClass = get_BoundStubClass(stub_type, ip_addr, profile['__call__'])
		# this ensures the stub will inherit from a stub class that's bound to the configuration
		parent_classes = (BoundStubClass, ) + parent_classes

	ServiceStub = type('ServiceStub', parent_classes, attrs)
	return ServiceStub()

def get_BoundStubClass(stub_type, ip_addr, configuration):
	
	# a class for stubs that are bound to a certain RPC
	class BoundStubClass(stub_type):
		def __init__(self):
			stub_type.__init__(self, worker_ip=ip_addr, endpoint_prefix=configuration['endpoint_prefix']+'/', rpc_name='__call__', comms_pattern=configuration['comms_pattern'])

	return BoundStubClass

def get_RPC_stub(ip_addr, configuration):
	comms_pattern = configuration['comms_pattern']
	stub = None

	if (comms_pattern == 'simplex'):
		stub = CoroSimplexStub(worker_ip=ip_addr, endpoint_prefix=configuration['endpoint_prefix']+'/', rpc_name='__call__')

	elif (comms_pattern == 'duplex'):
		stub = CoroDuplexStub(worker_ip=ip_addr, endpoint_prefix=configuration['endpoint_prefix']+'/', rpc_name='__call__')
	
	return stub

def get_RPC_stub_2(ip_addr, configuration):
	def return_fn(*args):
		return 'this is a test RPC stub'

	return return_fn

def get_worker_profile(ip_addr):
	url = 'http://'+str(ip_addr)+':'+str(comms_config.worker_port)+'/_get_profile'
	_, profile_str = GET(url)
	return deserialize(profile_str)

class RemoteWorker():

	def __init__(self, profile_or_ip):
		self.ip_addr = None
		self.rpcs = None

		profile = None

		# profile_or_ip is either an ip address or a worker profile, here we test to see which one
		try:
			profile_or_ip['ip_addr']
			# profile_or_ip has a member ip_addr, and so must be a profile
			profile = profile_or_ip

		except(TypeError):
			# profile_or_ip is an ip address
			profile = get_worker_profile(profile_or_ip)

		# sets up all the rfc stubs
		self.setup(profile)

	def setup(self, profile):
		self.ip_addr = profile['ip_addr']
		self.name = profile['name']

		# print('----------------------------------------------------------------')
		# print(profile['rpcs'].keys())
		# print(profile['rpcs']['fn.name'])
		# print('----------------------------------------------------------------')

		# this will need to be a lookup on a services key to a number of service profiles
		# self.setup_rpc_stubs(profile['rpcs'])
		self.rpcs = get_ServiceStub(self.ip_addr, endpoint_prefix='rpcs/', name='rpcs', profile=profile['rpcs'])

	def setup_rpc_stubs(self, rpcs_descs):
		rpcs = {}

		for rpc_desc in rpcs_descs:
			
			comms_pattern = rpc_desc['comms_pattern']
			name = rpc_desc['name']

			if (comms_pattern == 'simplex'):
				rpcs[name] = CoroSimplexStub(worker_ip=self.ip_addr, rpc_name=name)

			elif (comms_pattern == 'duplex'):
				rpcs[name] = CoroDuplexStub(worker_ip=self.ip_addr, rpc_name=name)

			else:
				raise BaseException('unrecognised communication pattern:'+str(comms_pattern))

		self.rpcs = SimpleNamespace(**rpcs)